@@ -1,5 +1,5 @@
 #!/usr/bin/env python3
-# pylint: disable=missing-docstring,not-an-iterable,too-many-locals,too-many-arguments,invalid-name,too-many-return-statements,too-many-branches,len-as-condition,too-many-statements,broad-except
+# pylint: disable=missing-docstring,not-an-iterable,too-many-locals,too-many-arguments,invalid-name,too-many-return-statements,too-many-branches,len-as-condition,too-many-statements,broad-except,unnecessary-lambda
 
 import datetime
 import pdb
@@ -54,19 +54,12 @@ INTEGER_TYPES = {'integer', 'smallint', 'bigint'}
 FLOAT_TYPES = {'real', 'double precision'}
 JSON_TYPES = {'json', 'jsonb'}
 
-#NB> numeric/decimal columns in postgres without a specified scale && precision
-#default to 'up to 131072 digits before the decimal point; up to 16383
-#digits after the decimal point'. For practical reasons, we are capping this at 74/38
-#  https://www.postgresql.org/docs/10/static/datatype-numeric.html#DATATYPE-NUMERIC-TABLE
-MAX_SCALE = 38
-MAX_PRECISION = 100
 
 def nullable_column(col_type, pk):
     if pk:
         return  [col_type]
     return ['null', col_type]
 
-
 def schema_for_column_datatype(c):
     schema = {}
     #remove any array notation from type information as we use a separate field for that
@@ -77,6 +70,7 @@ def schema_for_column_datatype(c):
         schema['minimum'] = -1 * (2**(c.numeric_precision - 1))
         schema['maximum'] = 2**(c.numeric_precision - 1) - 1
         return schema
+
     if data_type == 'money':
         schema['type'] = nullable_column('string', c.is_primary_key)
         return schema
@@ -111,23 +105,14 @@ def schema_for_column_datatype(c):
 
     if data_type == 'numeric':
         schema['type'] = nullable_column('number', c.is_primary_key)
-        if c.numeric_scale is None or c.numeric_scale > MAX_SCALE:
-            LOGGER.warning('capping decimal scale to 38.  THIS MAY CAUSE TRUNCATION')
-            scale = MAX_SCALE
-        else:
-            scale = c.numeric_scale
-
-        if c.numeric_precision is None or c.numeric_precision > MAX_PRECISION:
-            LOGGER.warning('capping decimal precision to 100.  THIS MAY CAUSE TRUNCATION')
-            precision = MAX_PRECISION
-        else:
-            precision = c.numeric_precision
+        scale = post_db.numeric_scale(c)
+        precision = post_db.numeric_precision(c)
 
         schema['exclusiveMaximum'] = True
-        schema['maximum'] = 10 ** (precision - scale)
-        schema['multipleOf'] = 10 ** (0 - scale)
+        schema['maximum'] = post_db.numeric_max(precision, scale)
+        schema['multipleOf'] = post_db.numeric_multiple_of(scale)
         schema['exclusiveMinimum'] = True
-        schema['minimum'] = -10 ** (precision - scale)
+        schema['minimum'] = post_db.numeric_min(precision, scale)
         return schema
 
     if data_type in {'time without time zone', 'time with time zone'}:
@@ -151,12 +136,15 @@ def schema_for_column_datatype(c):
 
     if data_type == 'character varying':
         schema['type'] = nullable_column('string', c.is_primary_key)
-        schema['maxLength'] = c.character_maximum_length
+        if c.character_maximum_length:
+            schema['maxLength'] = c.character_maximum_length
+
         return schema
 
     if data_type == 'character':
         schema['type'] = nullable_column('string', c.is_primary_key)
-        schema['maxLength'] = c.character_maximum_length
+        if c.character_maximum_length:
+            schema['maxLength'] = c.character_maximum_length
         return schema
 
     if data_type in {'cidr', 'inet', 'macaddr'}:
@@ -165,18 +153,74 @@ def schema_for_column_datatype(c):
 
     return schema
 
-def schema_for_column(c):
-    column_schema = {}
+def schema_name_for_numeric_array(precision, scale):
+    schema_name = 'sdc_recursive_decimal_{}_{}_array'.format(precision, scale)
+    return schema_name
 
+def schema_for_column(c):
     #NB> from the post postgres docs: The current implementation does not enforce the declared number of dimensions either.
     #these means we can say nothing about an array column. its items may be more arrays or primitive types like integers
     #and this can vary on a row by row basis
-    if c.is_array:
-        column_schema["type"] = ["null", "array"]
-        column_schema["items"] = {}
-        return Schema.from_dict(column_schema)
 
-    return Schema.from_dict(schema_for_column_datatype(c))
+    column_schema = {'type':["null", "array"]}
+    if not c.is_array:
+        return schema_for_column_datatype(c)
+
+    if c.sql_data_type == 'integer[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_integer_array'}
+    elif c.sql_data_type == 'bigint[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_integer_array'}
+    elif c.sql_data_type == 'bit[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_boolean_array'}
+    elif c.sql_data_type == 'boolean[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_boolean_array'}
+    elif c.sql_data_type == 'character varying[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'cidr[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'citext[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'date[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_timestamp_array'}
+    elif c.sql_data_type == 'numeric[]':
+        scale = post_db.numeric_scale(c)
+        precision = post_db.numeric_precision(c)
+        schema_name = schema_name_for_numeric_array(precision, scale)
+        column_schema['items'] = {'$ref': '#/definitions/{}'.format(schema_name)}
+    elif c.sql_data_type == 'double precision[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_number_array'}
+    elif c.sql_data_type == 'hstore[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_object_array'}
+    elif c.sql_data_type == 'inet[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'json[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'jsonb[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'mac[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'money[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'real[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_number_array'}
+    elif c.sql_data_type == 'smallint[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_integer_array'}
+    elif c.sql_data_type == 'text[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'timestamp without time zone[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_timestamp_array'}
+    elif c.sql_data_type == 'timestamp with time zone[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_timestamp_array'}
+    elif c.sql_data_type == 'time[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    elif c.sql_data_type == 'uuid[]':
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    else:
+        #custom datatypes like enums
+        column_schema['items'] = {'$ref': '#/definitions/sdc_recursive_string_array'}
+    return column_schema
+
+
 
 #typlen  -1  == variable length arrays
 #typelem != 0 points to subtypes. 23 in the case of arrays
@@ -221,6 +265,7 @@ JOIN pg_catalog.pg_namespace n
 LEFT OUTER JOIN pg_index as i
   ON a.attrelid = i.indrelid
  AND a.attnum = ANY(i.indkey)
+ AND i.indisprimary = true
 LEFT OUTER JOIN pg_type AS subpgt
   ON pgt.typelem = subpgt.oid
  AND pgt.typelem != 0
@@ -259,10 +304,38 @@ def write_sql_data_type_md(mdata, col_info):
 
     return mdata
 
+
+BASE_RECURSIVE_SCHEMAS = {'sdc_recursive_integer_array'   : {'type' : ['null', 'integer', 'array'], 'items' : {'$ref': '#/definitions/sdc_recursive_integer_array'}},
+                          'sdc_recursive_number_array'    : {'type' : ['null', 'number', 'array'], 'items'  : {'$ref': '#/definitions/sdc_recursive_number_array'}},
+                          'sdc_recursive_string_array'    : {'type' : ['null', 'string', 'array'], 'items'  : {'$ref': '#/definitions/sdc_recursive_string_array'}},
+                          'sdc_recursive_boolean_array'   : {'type' : ['null', 'boolean', 'array'], 'items' : {'$ref': '#/definitions/sdc_recursive_boolean_array'}},
+                          'sdc_recursive_timestamp_array' : {'type' : ['null', 'string', 'array'], 'format' : 'date-time', 'items' : {'$ref': '#/definitions/sdc_recursive_timestamp_array'}},
+                          'sdc_recursive_object_array'    : {'type' : ['null', 'object', 'array'], 'items' : {'$ref': '#/definitions/sdc_recursive_object_array'}}}
+
+def include_array_schemas(columns, schema):
+    schema['definitions'] = copy.deepcopy(BASE_RECURSIVE_SCHEMAS)
+
+
+    decimal_array_columns = [key for key, value in columns.items() if value.sql_data_type == 'numeric[]']
+    for c in decimal_array_columns:
+        scale = post_db.numeric_scale(columns[c])
+        precision = post_db.numeric_precision(columns[c])
+        schema_name = schema_name_for_numeric_array(precision, scale)
+        schema['definitions'][schema_name] = {'type' : ['null', 'number', 'array'],
+                                              'multipleOf': post_db.numeric_multiple_of(scale),
+                                              'exclusiveMaximum' : True,
+                                              'maximum' : post_db.numeric_max(precision, scale),
+                                              'exclusiveMinimum': True,
+                                              'minimum' : post_db.numeric_min(precision, scale),
+                                              'items' : {'$ref': '#/definitions/{}'.format(schema_name)}}
+
+    return schema
+
 def discover_columns(connection, table_info):
     entries = []
     for schema_name in table_info.keys():
         for table_name in table_info[schema_name].keys():
+
             mdata = {}
             columns = table_info[schema_name][table_name]['columns']
             table_pks = [col_name for col_name, col_info in columns.items() if col_info.is_primary_key]
@@ -278,10 +351,16 @@ def discover_columns(connection, table_info):
 
             column_schemas = {col_name : schema_for_column(col_info) for col_name, col_info in columns.items()}
 
-            schema = Schema(type='object', properties=column_schemas)
+            schema = {'type' : 'object',
+                      'properties': column_schemas,
+                      'definitions' : {}}
+
+            schema = include_array_schemas(columns, schema)
+
             for c_name in column_schemas.keys():
                 mdata = write_sql_data_type_md(mdata, columns[c_name])
-                if column_schemas[c_name].type is None:
+
+                if column_schemas[c_name].get('type') is None:
                     mdata = metadata.write(mdata, ('properties', c_name), 'inclusion', 'unsupported')
                     mdata = metadata.write(mdata, ('properties', c_name), 'selected-by-default', False)
                 elif table_info[schema_name][table_name]['columns'][c_name].is_primary_key:
@@ -291,20 +370,16 @@ def discover_columns(connection, table_info):
                     mdata = metadata.write(mdata, ('properties', c_name), 'inclusion', 'available')
                     mdata = metadata.write(mdata, ('properties', c_name), 'selected-by-default', True)
 
-            entry = CatalogEntry(
-                table=table_name,
-                stream=table_name,
-                metadata=metadata.to_list(mdata),
-                tap_stream_id=post_db.compute_tap_stream_id(database_name, schema_name, table_name),
-                schema=schema)
+            entry = {'table_name'    : table_name,
+                     'stream'        : table_name,
+                     'metadata'      : metadata.to_list(mdata),
+                     'tap_stream_id' : post_db.compute_tap_stream_id(database_name, schema_name, table_name),
+                     'schema'        : schema}
 
             entries.append(entry)
 
     return entries
 
-def dump_catalog(catalog):
-    catalog.dump()
-
 def discover_db(connection):
     table_info = produce_table_info(connection)
     db_streams = discover_columns(connection, table_info)
@@ -323,6 +398,9 @@ def attempt_connection_to_db(conn_config, dbname):
         LOGGER.warning('Unable to connect to %s. This maybe harmless if you have not desire to replicate from this database: "%s"', dbname, err)
         return False
 
+def dump_catalog(all_streams):
+    json.dump({'streams' : all_streams}, sys.stdout, indent=2)
+
 def do_discovery(conn_config):
     all_streams = []
 
@@ -353,34 +431,24 @@ def do_discovery(conn_config):
 
 
     if len(all_streams) == 0:
-        LOGGER.warning("0 tables were discovered across the entire cluster")
+        raise RuntimeError('0 tables were discovered across the entire cluster')
+
+    dump_catalog(all_streams)
+    return all_streams
 
-    cluster_catalog = Catalog(all_streams)
-    dump_catalog(cluster_catalog)
-    return cluster_catalog
 
 def should_sync_column(md_map, field_name):
-    #always sync replication_keys
-    if md_map.get((), {}).get('replication-key') == field_name:
-        return True
-
-    if md_map.get(('properties', field_name), {}).get('inclusion') == 'unsupported':
-        return False
-
-    if md_map.get(('properties', field_name), {}).get('selected'):
-        return True
-
-    if md_map.get(('properties', field_name), {}).get('inclusion') == 'automatic':
-        return True
-
-    return False
+    field_metadata = md_map.get(('properties', field_name), {})
+    return singer.should_sync_field(field_metadata.get('inclusion'),
+                                    field_metadata.get('selected'),
+                                    True)
 
 def is_selected_via_metadata(stream):
-    table_md = metadata.to_map(stream.metadata).get((), {})
+    table_md = metadata.to_map(stream['metadata']).get((), {})
     return table_md.get('selected')
 
 def do_sync_full_table(conn_config, stream, state, desired_columns, md_map):
-    LOGGER.info("Stream %s is using full_table replication", stream.tap_stream_id)
+    LOGGER.info("Stream %s is using full_table replication", stream['tap_stream_id'])
     sync_common.send_schema_message(stream, [])
     if md_map.get((), {}).get('is-view'):
         state = full_table.sync_view(conn_config, stream, state, desired_columns, md_map)
@@ -391,20 +459,16 @@ def do_sync_full_table(conn_config, stream, state, desired_columns, md_map):
 #Possible state keys: replication_key, replication_key_value, version
 def do_sync_incremental(conn_config, stream, state, desired_columns, md_map):
     replication_key = md_map.get((), {}).get('replication-key')
-    LOGGER.info("Stream %s is using incremental replication with replication key %s", stream.tap_stream_id, replication_key)
+    LOGGER.info("Stream %s is using incremental replication with replication key %s", stream['tap_stream_id'], replication_key)
 
-    stream_state = state.get('bookmarks', {}).get(stream.tap_stream_id)
+    stream_state = state.get('bookmarks', {}).get(stream['tap_stream_id'])
     illegal_bk_keys = set(stream_state.keys()).difference(set(['replication_key', 'replication_key_value', 'version', 'last_replication_method']))
     if len(illegal_bk_keys) != 0:
         raise Exception("invalid keys found in state: {}".format(illegal_bk_keys))
 
-    state = singer.write_bookmark(state, stream.tap_stream_id, 'replication_key', replication_key)
-    if md_map.get((), {}).get('is-view'):
-        pks = md_map.get(()).get('table-key-properties')
-    else:
-        pks = md_map.get(()).get('table-key-properties')
+    state = singer.write_bookmark(state, stream['tap_stream_id'], 'replication_key', replication_key)
 
-    sync_common.send_schema_message(stream, pks)
+    sync_common.send_schema_message(stream, [replication_key])
     state = incremental.sync_table(conn_config, stream, state, desired_columns, md_map)
 
     return state
@@ -429,106 +493,169 @@ def sync_method_for_streams(streams, state, default_replication_method):
     logical_streams = []
 
     for stream in streams:
-        stream_metadata = metadata.to_map(stream.metadata)
+        stream_metadata = metadata.to_map(stream['metadata'])
         replication_method = stream_metadata.get((), {}).get('replication-method', default_replication_method)
         replication_key = stream_metadata.get((), {}).get('replication-key')
 
-        state = clear_state_on_replication_change(state, stream.tap_stream_id, replication_key, replication_method)
+        state = clear_state_on_replication_change(state, stream['tap_stream_id'], replication_key, replication_method)
 
         if replication_method not in set(['LOG_BASED', 'FULL_TABLE', 'INCREMENTAL']):
             raise Exception("Unrecognized replication_method {}".format(replication_method))
 
-        md_map = metadata.to_map(stream.metadata)
-        desired_columns = [c for c in stream.schema.properties.keys() if should_sync_column(md_map, c)]
+        md_map = metadata.to_map(stream['metadata'])
+        desired_columns = [c for c in stream['schema']['properties'].keys() if should_sync_column(md_map, c)]
         desired_columns.sort()
 
         if len(desired_columns) == 0:
-            LOGGER.warning('There are no columns selected for stream %s, skipping it', stream.tap_stream_id)
+            LOGGER.warning('There are no columns selected for stream %s, skipping it', stream['tap_stream_id'])
             continue
 
         if replication_method == 'LOG_BASED' and stream_metadata.get((), {}).get('is-view'):
-            raise Exception('Logical Replication is NOT supported for views. Please change the replication method for {}'.format(stream.tap_stream_id))
+            raise Exception('Logical Replication is NOT supported for views. Please change the replication method for {}'.format(stream['tap_stream_id']))
 
         if replication_method == 'FULL_TABLE':
-            lookup[stream.tap_stream_id] = 'full'
+            lookup[stream['tap_stream_id']] = 'full'
             traditional_steams.append(stream)
         elif replication_method == 'INCREMENTAL':
-            lookup[stream.tap_stream_id] = 'incremental'
+            lookup[stream['tap_stream_id']] = 'incremental'
             traditional_steams.append(stream)
 
-        elif get_bookmark(state, stream.tap_stream_id, 'xmin') and get_bookmark(state, stream.tap_stream_id, 'lsn'):
+        elif get_bookmark(state, stream['tap_stream_id'], 'xmin') and get_bookmark(state, stream['tap_stream_id'], 'lsn'):
             #finishing previously interrupted full-table (first stage of logical replication)
-            lookup[stream.tap_stream_id] = 'logical_initial_interrupted'
+            lookup[stream['tap_stream_id']] = 'logical_initial_interrupted'
             traditional_steams.append(stream)
 
         #inconsistent state
-        elif get_bookmark(state, stream.tap_stream_id, 'xmin') and not get_bookmark(state, stream.tap_stream_id, 'lsn'):
+        elif get_bookmark(state, stream['tap_stream_id'], 'xmin') and not get_bookmark(state, stream['tap_stream_id'], 'lsn'):
             raise Exception("Xmin found(%s) in state implying full-table replication but no lsn is present")
 
-        elif not get_bookmark(state, stream.tap_stream_id, 'xmin') and not get_bookmark(state, stream.tap_stream_id, 'lsn'):
+        elif not get_bookmark(state, stream['tap_stream_id'], 'xmin') and not get_bookmark(state, stream['tap_stream_id'], 'lsn'):
             #initial full-table phase of logical replication
-            lookup[stream.tap_stream_id] = 'logical_initial'
+            lookup[stream['tap_stream_id']] = 'logical_initial'
             traditional_steams.append(stream)
 
         else: #no xmin but we have an lsn
             #initial stage of logical replication(full-table) has been completed. moving onto pure logical replication
-            lookup[stream.tap_stream_id] = 'pure_logical'
+            lookup[stream['tap_stream_id']] = 'pure_logical'
             logical_streams.append(stream)
 
     return lookup, traditional_steams, logical_streams
 
-def sync_traditional_stream(conn_config, stream, state, sync_method):
-    LOGGER.info("Beginning sync of stream(%s) with sync method(%s)", stream.tap_stream_id, sync_method)
-    md_map = metadata.to_map(stream.metadata)
+def sync_traditional_stream(conn_config, stream, state, sync_method, end_lsn):
+    LOGGER.info("Beginning sync of stream(%s) with sync method(%s)", stream['tap_stream_id'], sync_method)
+    md_map = metadata.to_map(stream['metadata'])
     conn_config['dbname'] = md_map.get(()).get('database-name')
-    desired_columns = [c for c in stream.schema.properties.keys() if should_sync_column(md_map, c)]
+    desired_columns = [c for c in stream['schema']['properties'].keys() if should_sync_column(md_map, c)]
     desired_columns.sort()
 
     if len(desired_columns) == 0:
-        LOGGER.warning('There are no columns selected for stream %s, skipping it', stream.tap_stream_id)
+        LOGGER.warning('There are no columns selected for stream %s, skipping it', stream['tap_stream_id'])
         return state
 
+    register_type_adapters(conn_config)
+
     if sync_method == 'full':
-        state = singer.set_currently_syncing(state, stream.tap_stream_id)
+        state = singer.set_currently_syncing(state, stream['tap_stream_id'])
         state = do_sync_full_table(conn_config, stream, state, desired_columns, md_map)
     elif sync_method == 'incremental':
-        state = singer.set_currently_syncing(state, stream.tap_stream_id)
+        state = singer.set_currently_syncing(state, stream['tap_stream_id'])
         state = do_sync_incremental(conn_config, stream, state, desired_columns, md_map)
     elif sync_method == 'logical_initial':
-        state = singer.set_currently_syncing(state, stream.tap_stream_id)
-        end_lsn = logical_replication.fetch_current_lsn(conn_config)
+        state = singer.set_currently_syncing(state, stream['tap_stream_id'])
         LOGGER.info("Performing initial full table sync")
-        state = singer.write_bookmark(state, stream.tap_stream_id, 'lsn', end_lsn)
+        state = singer.write_bookmark(state, stream['tap_stream_id'], 'lsn', end_lsn)
 
         sync_common.send_schema_message(stream, [])
         state = full_table.sync_table(conn_config, stream, state, desired_columns, md_map)
-        state = singer.write_bookmark(state, stream.tap_stream_id, 'xmin', None)
+        state = singer.write_bookmark(state, stream['tap_stream_id'], 'xmin', None)
     elif sync_method == 'logical_initial_interrupted':
-        state = singer.set_currently_syncing(state, stream.tap_stream_id)
+        state = singer.set_currently_syncing(state, stream['tap_stream_id'])
         LOGGER.info("Initial stage of full table sync was interrupted. resuming...")
         sync_common.send_schema_message(stream, [])
         state = full_table.sync_table(conn_config, stream, state, desired_columns, md_map)
     else:
-        raise Exception("unknown sync method {} for stream {}".format(sync_method, stream.tap_stream_id))
+        raise Exception("unknown sync method {} for stream {}".format(sync_method, stream['tap_stream_id']))
 
     state = singer.set_currently_syncing(state, None)
     singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))
     return state
 
-def sync_logical_streams(conn_config, logical_streams, state):
+def sync_logical_streams(conn_config, logical_streams, state, end_lsn):
     if logical_streams:
-        LOGGER.info("Pure Logical Replication upto lsn %s for (%s)", logical_replication.fetch_current_lsn(conn_config), list(map(lambda s: s.tap_stream_id, logical_streams)))
+        LOGGER.info("Pure Logical Replication upto lsn %s for (%s)", end_lsn, list(map(lambda s: s['tap_stream_id'], logical_streams)))
         logical_streams = list(map(logical_replication.add_automatic_properties, logical_streams))
 
-        state = logical_replication.sync_tables(conn_config, logical_streams, state)
+        state = logical_replication.sync_tables(conn_config, logical_streams, state, end_lsn)
 
     return state
 
+
+def register_type_adapters(conn_config):
+    with post_db.open_connection(conn_config) as conn:
+        with conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
+            #citext[]
+            cur.execute("SELECT typarray FROM pg_type where typname = 'citext'")
+            citext_array_oid = cur.fetchone()
+            if citext_array_oid:
+                psycopg2.extensions.register_type(
+                    psycopg2.extensions.new_array_type(
+                        (citext_array_oid[0],), 'CITEXT[]', psycopg2.STRING))
+
+            #bit[]
+            cur.execute("SELECT typarray FROM pg_type where typname = 'bit'")
+            bit_array_oid = cur.fetchone()[0]
+            psycopg2.extensions.register_type(
+                psycopg2.extensions.new_array_type(
+                    (bit_array_oid,), 'BIT[]', psycopg2.STRING))
+
+
+            #UUID[]
+            cur.execute("SELECT typarray FROM pg_type where typname = 'uuid'")
+            uuid_array_oid = cur.fetchone()[0]
+            psycopg2.extensions.register_type(
+                psycopg2.extensions.new_array_type(
+                    (uuid_array_oid,), 'UUID[]', psycopg2.STRING))
+
+            #money[]
+            cur.execute("SELECT typarray FROM pg_type where typname = 'money'")
+            money_array_oid = cur.fetchone()[0]
+            psycopg2.extensions.register_type(
+                psycopg2.extensions.new_array_type(
+                    (money_array_oid,), 'MONEY[]', psycopg2.STRING))
+
+            #json and jsbon
+            psycopg2.extras.register_default_json(loads=lambda x: str(x))
+            psycopg2.extras.register_default_jsonb(loads=lambda x: str(x))
+
+            #enum[]'s
+            cur.execute("SELECT distinct(t.typarray) FROM pg_type t JOIN pg_enum e ON t.oid = e.enumtypid")
+            for oid in cur.fetchall():
+                enum_oid = oid[0]
+                psycopg2.extensions.register_type(
+                    psycopg2.extensions.new_array_type(
+                        (enum_oid,), 'ENUM_{}[]'.format(enum_oid), psycopg2.STRING))
+
+
+def any_logical_streams(streams, default_replication_method):
+    for stream in streams:
+        stream_metadata = metadata.to_map(stream['metadata'])
+        replication_method = stream_metadata.get((), {}).get('replication-method', default_replication_method)
+        if replication_method == 'LOG_BASED':
+            return True
+
+    return False
+
 def do_sync(conn_config, catalog, default_replication_method, state):
     currently_syncing = singer.get_currently_syncing(state)
-    streams = list(filter(is_selected_via_metadata, catalog.streams))
-    streams.sort(key=lambda s: s.tap_stream_id)
-    LOGGER.info("Selected streams: %s ", list(map(lambda s: s.tap_stream_id, streams)))
+    streams = list(filter(is_selected_via_metadata, catalog['streams']))
+    streams.sort(key=lambda s: s['tap_stream_id'])
+    LOGGER.info("Selected streams: %s ", list(map(lambda s: s['tap_stream_id'], streams)))
+    if any_logical_streams(streams, default_replication_method):
+        LOGGER.info("Use of logical replication requires fetching an lsn...")
+        end_lsn = logical_replication.fetch_current_lsn(conn_config)
+        LOGGER.info("End LSN: %s ", end_lsn)
+    else:
+        end_lsn = None
 
     sync_method_lookup, traditional_streams, logical_streams = sync_method_for_streams(streams, state, default_replication_method)
     #{"chickens" : "full_stream", "cows" : "logical_initial_interrupted_streams", "turkeys": "logical_replication"}
@@ -536,21 +663,21 @@ def do_sync(conn_config, catalog, default_replication_method, state):
 
     if currently_syncing:
         LOGGER.info("found currently_syncing: %s", currently_syncing)
-        currently_syncing_stream = list(filter(lambda s: s.tap_stream_id == currently_syncing, traditional_streams))
+        currently_syncing_stream = list(filter(lambda s: s['tap_stream_id'] == currently_syncing, traditional_streams))
         if currently_syncing_stream is None:
-            LOGGER.warning("unable to locate currently_syncing(%s) amongst selected traditional streams(%s). will ignore", currently_syncing, list(map(lambda s: s.tap_stream_id, traditional_streams)))
-        other_streams = list(filter(lambda s: s.tap_stream_id != currently_syncing, traditional_streams))
+            LOGGER.warning("unable to locate currently_syncing(%s) amongst selected traditional streams(%s). will ignore", currently_syncing, list(map(lambda s: s['tap_stream_id'], traditional_streams)))
+        other_streams = list(filter(lambda s: s['tap_stream_id'] != currently_syncing, traditional_streams))
         traditional_streams = currently_syncing_stream + other_streams
     else:
         LOGGER.info("No currently_syncing found")
 
     for stream in traditional_streams:
-        state = sync_traditional_stream(conn_config, stream, state, sync_method_lookup[stream.tap_stream_id])
+        state = sync_traditional_stream(conn_config, stream, state, sync_method_lookup[stream['tap_stream_id']], end_lsn)
 
-    logical_streams.sort(key=lambda s: metadata.to_map(s.metadata).get(()).get('database-name'))
-    for dbname, streams in itertools.groupby(logical_streams, lambda s: metadata.to_map(s.metadata).get(()).get('database-name')):
+    logical_streams.sort(key=lambda s: metadata.to_map(s['metadata']).get(()).get('database-name'))
+    for dbname, streams in itertools.groupby(logical_streams, lambda s: metadata.to_map(s['metadata']).get(()).get('database-name')):
         conn_config['dbname'] = dbname
-        state = sync_logical_streams(conn_config, list(streams), state)
+        state = sync_logical_streams(conn_config, list(streams), state, end_lsn)
     return state
 
 def main_impl():
@@ -571,9 +698,9 @@ def main_impl():
 
     if args.discover:
         do_discovery(conn_config)
-    elif args.catalog:
+    elif args.properties:
         state = args.state
-        do_sync(conn_config, args.catalog, args.config.get('default_replication_method'), state)
+        do_sync(conn_config, args.properties, args.config.get('default_replication_method'), state)
     else:
         LOGGER.info("No properties were selected")
 
